@@ -5,9 +5,9 @@
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
  * You may obtain a copy of the License at
- * 
- * 	http://www.apache.org/licenses/LICENSE-2.0
- * 
+ *
+ *	http://www.apache.org/licenses/LICENSE-2.0
+ *
  * Unless required by applicable law or agreed to in writing, software
  * distributed under the License is distributed on an "AS IS" BASIS,
  * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
@@ -19,6 +19,7 @@
 
 #include <unistd.h>		/* getegid */
 #include <stdio.h>		/* fgetgrent */
+#include <sys/stat.h>		/* umask */
 #include <sys/types.h>		/* fgetgrent */
 #include <grp.h>		/* fgetgrent */
 
@@ -109,6 +110,9 @@ typedef struct ft_conf_t
 {
     apr_off_t minsize;
     apr_off_t excess_size;	/* Switch off mmap behavior */
+#if HAVE_PUZZLE
+    double threshold;
+#endif
     apr_pool_t *pool;		/* Always needed somewhere ;) */
     napr_heap_t *heap;		/* Will holds the files */
     napr_hash_t *sizes;		/* will holds the sizes hashed with http://www.burtleburtle.net/bob/hash/integer.html */
@@ -126,6 +130,12 @@ typedef struct ft_conf_t
     char sep;
 } ft_conf_t;
 
+struct stats
+{
+    struct stats const *parent;
+    apr_finfo_t stat;
+};
+
 static int ft_file_cmp(const void *param1, const void *param2)
 {
     const ft_file_t *file1 = param1;
@@ -192,7 +202,7 @@ static void ft_hash_add_ignore_list(napr_hash_t *hash, const char *file_list)
     pool = napr_hash_pool_get(hash);
     filename = file_list;
     do {
-	end = strchr(filename, 'c');
+	end = strchr(filename, ',');
 	if (NULL != end) {
 	    tmp = apr_pstrndup(pool, filename, end - filename);
 	}
@@ -206,7 +216,7 @@ static void ft_hash_add_ignore_list(napr_hash_t *hash, const char *file_list)
     } while ((NULL != end) && ('\0' != *filename));
 }
 
-/** 
+/**
  * The function used to add recursively or not file and dirs.
  * @param conf Configuration structure.
  * @param filename name of a file or directory to add to the list of twinchecker.
@@ -214,7 +224,7 @@ static void ft_hash_add_ignore_list(napr_hash_t *hash, const char *file_list)
  * @return APR_SUCCESS if no error occured.
  */
 #define MATCH_VECTOR_SIZE 210
-static apr_status_t ft_conf_add_file(ft_conf_t *conf, const char *filename, apr_pool_t *gc_pool)
+static apr_status_t ft_conf_add_file(ft_conf_t *conf, const char *filename, apr_pool_t *gc_pool, struct stats const *stats)
 {
     int ovector[MATCH_VECTOR_SIZE];
     char errbuf[128];
@@ -222,7 +232,7 @@ static apr_status_t ft_conf_add_file(ft_conf_t *conf, const char *filename, apr_
     apr_dir_t *dir;
     apr_int32_t statmask =
 	APR_FINFO_SIZE | APR_FINFO_TYPE | APR_FINFO_USER | APR_FINFO_GROUP | APR_FINFO_UPROT | APR_FINFO_GPROT;
-    apr_size_t fname_len;
+    apr_size_t fname_len = 0;
     apr_uint32_t hash_value;
     apr_status_t status;
     int rc;
@@ -232,6 +242,15 @@ static apr_status_t ft_conf_add_file(ft_conf_t *conf, const char *filename, apr_
 	statmask |= APR_FINFO_LINK;
 
     if (APR_SUCCESS != (status = apr_stat(&finfo, filename, statmask, gc_pool))) {
+	if (is_option_set(conf->mask, OPTION_FSYML)) {
+	    statmask ^= APR_FINFO_LINK;
+	    if ((APR_SUCCESS == apr_stat(&finfo, filename, statmask, gc_pool)) && (finfo.filetype & APR_LNK)) {
+		if (is_option_set(conf->mask, OPTION_VERBO))
+		    fprintf(stderr, "Skipping : [%s] (broken link)\n", filename);
+		return APR_SUCCESS;
+	    }
+	}
+
 	DEBUG_ERR("error calling apr_stat on filename %s : %s", filename, apr_strerror(status, errbuf, 128));
 	return status;
     }
@@ -293,6 +312,9 @@ static apr_status_t ft_conf_add_file(ft_conf_t *conf, const char *filename, apr_
 	    /* Check if it has to be ignored */
 	    char *fullname;
 	    apr_size_t fullname_len;
+	    /* for recursive loop detection */
+	    struct stats child;
+	    struct stats const *ancestor;
 
 	    if (NULL != napr_hash_search(conf->ig_files, finfo.name, strlen(finfo.name), NULL))
 		continue;
@@ -311,7 +333,20 @@ static apr_status_t ft_conf_add_file(ft_conf_t *conf, const char *filename, apr_
 		&& (0 > (rc = pcre_exec(conf->wl_regex, NULL, fullname, fullname_len, 0, 0, ovector, MATCH_VECTOR_SIZE))))
 		continue;
 
-	    status = ft_conf_add_file(conf, fullname, gc_pool);
+	    if (stats) {
+		if (stats->stat.inode)
+		    for (ancestor = stats; (ancestor = ancestor->parent) != 0;)
+			if (ancestor->stat.inode == stats->stat.inode && ancestor->stat.device == stats->stat.device) {
+			    if (is_option_set(conf->mask, OPTION_VERBO))
+				fprintf(stderr, "Warning: %s: recursive directory loop\n", filename);
+			    /* skip it */
+			    return APR_SUCCESS;
+			}
+	    }
+	    child.parent = stats;
+	    child.stat = finfo;
+
+	    status = ft_conf_add_file(conf, fullname, gc_pool, &child);
 
 	    if (APR_SUCCESS != status) {
 		DEBUG_ERR("error recursively calling ft_conf_add_file: %s", apr_strerror(status, errbuf, 128));
@@ -518,6 +553,8 @@ static char *ft_untar_file(ft_file_t *file, apr_pool_t *p)
 	}
 
 	if (!strcmp(file->subpath, archive_entry_pathname(entry))) {
+	    mode_t current_mode = umask(S_IRWXG | S_IRWXO);
+
 	    /*
 	     * All I want is only a temporary filename, but gcc outputs me an
 	     * ugly warning if I use tempnam...
@@ -529,6 +566,7 @@ static char *ft_untar_file(ft_file_t *file, apr_pool_t *p)
 		DEBUG_ERR("error creating tmpfile %s", tmpfile);
 		return NULL;
 	    }
+	    umask(current_mode);
 	    close(rv);
 
 	    archive_entry_copy_pathname(entry, tmpfile);
@@ -731,7 +769,7 @@ static apr_status_t ft_conf_image_twin_report(ft_conf_t *conf)
 		continue;
 
 	    d = puzzle_vector_normalized_distance(&context, &(file->cvec), &(file_cmp->cvec), 0);
-	    if (d < PUZZLE_CVEC_SIMILARITY_LOWER_THRESHOLD) {
+	    if (d < conf->threshold) {
 		if (!already_printed) {
 		    printf("%s%c", file->path, conf->sep);
 		    already_printed = 1;
@@ -766,6 +804,7 @@ static apr_status_t ft_conf_twin_report(ft_conf_t *conf)
     int rv;
     apr_status_t status;
     unsigned char already_printed;
+    apr_uint32_t chksum_array_sz = 0U;
 
     if (is_option_set(conf->mask, OPTION_VERBO))
 	fprintf(stderr, "Reporting duplicate files:\n");
@@ -776,7 +815,8 @@ static apr_status_t ft_conf_twin_report(ft_conf_t *conf)
 
 	old_size = file->size;
 	if (NULL != (fsize = napr_hash_search(conf->sizes, &file->size, 1, &hash_value))) {
-	    qsort(fsize->chksum_array, fsize->nb_files, sizeof(ft_chksum_t), chksum_cmp);
+	    chksum_array_sz = MIN(fsize->nb_files, fsize->nb_checksumed);
+	    qsort(fsize->chksum_array, chksum_array_sz, sizeof(ft_chksum_t), chksum_cmp);
 	    for (i = 0; i < fsize->nb_files; i++) {
 		if (NULL == fsize->chksum_array[i].file)
 		    continue;
@@ -951,7 +991,7 @@ static apr_status_t fill_gids_ht(const char *username, napr_hash_t *gids, apr_po
     apr_uint32_t hash_value;
     int i, nb_gid;
 
-    nb_gid = getgroups(sizeof(list), list);
+    nb_gid = getgroups(sizeof(list) / sizeof(gid_t), list);
     if (nb_gid < 0) {
 	DEBUG_ERR("error calling getgroups()");
 	return APR_EGENERAL;
@@ -962,13 +1002,14 @@ static apr_status_t fill_gids_ht(const char *username, napr_hash_t *gids, apr_po
      * is included in the returned list.  (Thus, an application should also
      * call getegid(2) and add or remove the resulting value.)
      */
-    if (nb_gid < sizeof(list)) {
+    if (nb_gid < (sizeof(list) / sizeof(gid_t))) {
 	list[nb_gid] = getegid();
 	nb_gid++;
     }
 
     for (i = 0; i < nb_gid; i++) {
-	if (NULL == (gid = napr_hash_search(gids, &(list[i]), 1, &hash_value))) {
+	gid = napr_hash_search(gids, &(list[i]), 1, &hash_value);
+	if (NULL == gid) {
 	    gid = apr_palloc(p, sizeof(struct ft_gid_t));
 	    gid->val = list[i];
 	    napr_hash_set(gids, gid, hash_value);
@@ -988,6 +1029,8 @@ int main(int argc, const char **argv)
 	{"help", 'h', FALSE, "\t\tdisplay usage."},
 #if HAVE_PUZZLE
 	{"image-cmp", 'I', FALSE, "\twill run ftwin in image cmp mode (using libpuzzle)."},
+	{"image-threshold", 'T', TRUE,
+	 "will change the image similarity threshold\n\t\t\t\t (default is [1], accepted [2/3/4/5])."},
 #endif
 	{"ignore-list", 'i', TRUE, "\tcomma-separated list of file names to ignore."},
 	{"minimal-length", 'm', TRUE, "minimum size of file to process."},
@@ -1049,7 +1092,10 @@ int main(int argc, const char **argv)
     conf.minsize = 0;
     conf.sep = '\n';
     conf.excess_size = 50 * 1024 * 1024;
-    conf.mask = 0x00;
+    conf.mask = 0x0000;
+#if HAVE_ARCHIVE
+    conf.threshold = PUZZLE_CVEC_SIMILARITY_LOWER_THRESHOLD;
+#endif
 
     while (APR_SUCCESS == (status = apr_getopt_long(os, opt_option, &optch, &optarg))) {
 	switch (optch) {
@@ -1075,7 +1121,31 @@ int main(int argc, const char **argv)
 	case 'I':
 	    set_option(&conf.mask, OPTION_ICASE, 1);
 	    set_option(&conf.mask, OPTION_PUZZL, 1);
-	    wregex = apr_pstrdup(pool, ".*\\.(gif|png|jpg)$");
+	    wregex = apr_pstrdup(pool, ".*\\.(gif|png|jpe?g)$");
+	    break;
+	case 'T':
+	    switch (*optarg) {
+	    case '1':
+		conf.threshold = PUZZLE_CVEC_SIMILARITY_LOWER_THRESHOLD;
+		break;
+	    case '2':
+		conf.threshold = PUZZLE_CVEC_SIMILARITY_LOW_THRESHOLD;
+		break;
+	    case '3':
+		conf.threshold = 0.5;
+		break;
+	    case '4':
+		conf.threshold = PUZZLE_CVEC_SIMILARITY_THRESHOLD;
+		break;
+	    case '5':
+		conf.threshold = PUZZLE_CVEC_SIMILARITY_HIGH_THRESHOLD;
+		break;
+	    default:
+		usage(argv[0], opt_option);
+		DEBUG_ERR("invalid threshold: %s", optarg);
+		apr_terminate();
+		return -1;
+	    }
 	    break;
 #endif
 	case 'm':
@@ -1174,7 +1244,7 @@ int main(int argc, const char **argv)
 	return -1;
     }
     for (i = os->ind; i < argc; i++) {
-	if (APR_SUCCESS != (status = ft_conf_add_file(&conf, argv[i], gc_pool))) {
+	if (APR_SUCCESS != (status = ft_conf_add_file(&conf, argv[i], gc_pool, NULL))) {
 	    DEBUG_ERR("error calling ft_conf_add_file: %s", apr_strerror(status, errbuf, 128));
 	    apr_terminate();
 	    return -1;
